/*==============================================================================


   Purpose:    low level routines for Alpha/Cray T3E


   Author:     Rudolf Berrendorf
               University of Applied Sciences Bonn-Rhein-Sieg
               Computer Science Department
               53754 Sankt Augustin
               rudolf.berrendorf@fh-bonn-rhein-sieg.de


   last mod:   23.08.2000


   DISCLAIMER

                  PCL - The Performance Counter Library


   Copyright (C) 2000
   Forschungszentrum Juelich GmbH, Federal Republic of Germany.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:

   Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

     - Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.

     - Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.

     - Any publications that result from the use of this software shall
       reasonably refer to the Research Centre's development.

     - All advertising materials mentioning features or use of this software
       must display the following acknowledgement:

           This product includes software developed by Forschungszentrum
           Juelich GmbH, Federal Republic of Germany.

     - Forschungszentrum Juelich GmbH is not obligated to provide the user with
       any support, consulting, training or assistance of any kind with regard
       to the use, operation and performance of this software or to provide
       the user with any updates, revisions or new versions.


   THIS SOFTWARE IS PROVIDED BY FORSCHUNGSZENTRUM JUELICH GMBH "AS IS" AND ANY
   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
   DISCLAIMED. IN NO EVENT SHALL FORSCHUNGSZENTRUM JUELICH GMBH BE LIABLE FOR
   ANY SPECIAL, DIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
   RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
   CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE ACCESS, USE OR PERFORMANCE OF THIS SOFTWARE.
  
==============================================================================*/

#include "../../defines.h"
#include "../../pcl.h"
#include "hw_driver_defs.h"
#include "../../pcl_hw_driver.h"
#include "../../hw_driver_global_defs.h"
#include "../../state_descr.h"

#include <mpp/globals.h>
#include <unistd.h>
#include "perfctr.h"

/*                                                                            */
/*----------------------------------------------------------------------------*/
/* constants                                                                  */

/* define the following macro if your T3E system might do a process migration */
/* ??? I am not sure if this check is still needed. */
/*#define CHECK_MIGRATION*/


/*                                                                            */
/*----------------------------------------------------------------------------*/
/* local variables                                                            */

/* entries are control codes for hardware performance counter.
   sel0-sel2 contain event selectors for counter0 - counter2. For counter1
   and counter2, CBOX events are decoded as event selectors larger than 0x0e.
   The CBOX value is then computed as (value - 0x0f).
*/
/* invalid entry-data if event is not available */
#define PCL_INVALID    0, {{{-1,-1,-1,-1},{0,0,0,0}}}

static const PCL_event_descr_t event_table_ev5[PCL_MAX_EVENT] =
{
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L1CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L1CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L1CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L1CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L1CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L1DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L1DCACHE_WRITE */
  {PCL_AVAIL,NOT_COMP,{{{-1,0x0E,-1,-1},{OP_NO,OP_ADD,OP_NO,OP_NO}}}}, /* PCL_L1DCACHE_READWRITE */
  {PCL_AVAIL,COMPOSED,{{{-1,0x0E,0x06,-1},{-1,OP_ADD,OP_SUB,-1}}}},/* PCL_L1DCACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,0x06,-1},{-1,-1,OP_ADD,-1}}}}, /* PCL_L1DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L1ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L1ICACHE_WRITE */
  {PCL_AVAIL,NOT_COMP,{{{-1,0x0D,-1,-1},{-1,OP_ADD,-1,-1}}}}, /* PCL_L1ICACHE_READWRITE */
  {PCL_AVAIL,COMPOSED,{{{-1,0x0D,0x04,-1},{-1,OP_ADD,OP_SUB,-1}}}},/* PCL_L1ICACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,0x04,-1},{-1,-1,OP_ADD,-1}}}}, /* PCL_L1ICACHE_MISS */
  {PCL_AVAIL,NOT_COMP,{{{-1,0x10,-1,-1},{-1,OP_ADD,-1,-1}}}},/* PCL_L2CACHE_READ */
  {PCL_AVAIL,NOT_COMP,{{{-1,0x11,-1,-1},{-1,OP_ADD,-1,-1}}}},/* PCL_L2CACHE_WRITE */
  {PCL_AVAIL,NOT_COMP,{{{-1,0x0F,-1,-1},{-1,OP_ADD,-1,-1}}}}, /* PCL_L2CACHE_READWRITE */
  {PCL_AVAIL,COMPOSED,{{{-1,0x0F,0x0F,-1},{-1,OP_ADD,OP_SUB,-1}}}},/* PCL_L2CACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,0x0F,-1},{-1,-1,OP_ADD,-1}}}}, /* PCL_L2CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L2DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L2DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L2DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L2DCACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L2DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L2ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L2ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L2ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L2ICACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L2ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_TLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_TLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_ITLB_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,0x05,-1},{-1,-1,OP_ADD,-1}}}}, /* PCL_ITLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_DTLB_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,0x07,-1},{-1,-1,OP_ADD,-1}}}}, /* PCL_DTLB_MISS */
  {PCL_AVAIL,NOT_COMP,{{{0x0,-1,0x0c,-1},{OP_ADD,-1,OP_ADD,-1}}}}, /* PCL_CYCLES */
  /* special case with dummy entry */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_ELAPSED_CYCLES */
  {PCL_AVAIL,NOT_COMP,{{{-1,0x09,-1,-1},{-1,OP_ADD,-1,-1}}}}, /* PCL_INTEGER_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{-1,0x0a,-1,-1},{-1,OP_ADD,-1,-1}}}}, /* PCL_FP_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{-1,0x0b,-1,-1},{-1,OP_ADD,-1,-1}}}}, /* PCL_LOAD_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{-1,0x0c,-1,-1},{-1,OP_ADD,-1,-1}}}}, /* PCL_STORE_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_LOADSTORE_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{0x01,-1,-1,-1},{OP_ADD,-1,-1,-1}}}}, /* PCL_INSTR */
  {PCL_AVAIL, PCL_INVALID},                      /* PCL_JUMP_SUCCESS */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,0x02,-1},{-1,-1,OP_ADD,-1}}}}, /* PCL_JUMP_UNSUCCESS */
  {PCL_AVAIL, PCL_INVALID},                      /* PCL_JUMP */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,0x0e,-1},{-1,-1,OP_ADD,-1}}}}, /* PCL_ATOMIC_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_ATOMIC_UNSUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_ATOMIC */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_STALL_INTEGER */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_STALL_FP */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_STALL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_STALL_LOAD */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_STALL_STORE */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_STALL */
  {PCL_AVAIL,COMPOSED,{{{-1,0xa,0xc,-1},{-1,OP_FP_2|OP_FP_STORE,OP_FP_2|OP_FP_DIV_MHZ,-1}}}}, /* PCL_MFLOPS */
  {PCL_AVAIL,COMPOSED,{{{0x1,-1,0xc,-1},{OP_FP_2|OP_FP_STORE,-1,OP_FP_2|OP_FP_DIV,-1}}}}, /* PCL_IPC */
  {PCL_AVAIL,COMPOSED,{{{-1,0xe,0x6,-1},{-1,OP_FP_1|OP_FP_2|OP_FP_REZ,OP_FP_2|OP_FP_MUL,-1}}}}, /* PCL_L1DCACHE_MISSRATE */
  {PCL_AVAIL,COMPOSED,{{{-1,0x0F,0x0F,-1},{-1,OP_FP_1|OP_FP_2|OP_FP_REZ,OP_FP_2|OP_FP_MUL,-1}}}}, /* PCL_L2DCACHE_MISSRATE */
  {PCL_UNAVAIL, PCL_INVALID}                     /* PCL_MEM_FP_RATIO */
};


/*                                                                            */
/*============================================================================*/
/* initialize processor number                                                */


int
PCL_driver_open(
		PCL_DESCR_TYPE d /* I: descriptor */
		)

{
  PCL_state_descr *descr;
  HW_state_descr *hw_descr;


  /* get state descriptor */
  descr = d;
  hw_descr = &descr->hw_state_descr;

  /* initialize global data */
  descr->event_table = event_table_ev5;
  descr->allowed_count_mode = PCL_MODE_USER_SYSTEM;

#if defined(CHECK_MIGRATION)
  /* get processor number we are running on */
  hw_descr->processornumber = PCL_processor_number();
#endif

  /* everything ok */
  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/* Starts hardware counters.                                                  */


int
PCL_driver_start(
		 PCL_DESCR_TYPE d, /* I: descriptor */
		 int max_counter,
		 PCL_BIT_MASK_TYPE counter_used_mask,
		 PCL_DRIVER_COMMAND_TYPE *commands,
		 unsigned int count_mode
		 )

{
  int sel0, sel1, sel2;	       /* counter selectors */
  unsigned int mode;
  PCL_BIT_MASK_TYPE used_mask, tmp;
  PCL_state_descr *descr;
  HW_state_descr *hw_descr;


  /* get state descriptor */
  descr = d;
  hw_descr = &descr->hw_state_descr;

  /* set up used mask */
  used_mask = counter_used_mask & PCL_mask[max_counter-1];

  /* initialize selectors with default count events (cycles,
     non_issue_cycles, long_stalls) */
  sel0 = sel1 = sel2 = 0;

  /* Extract selected events */
  if(IS_USED(used_mask, 0))
    sel0 = commands[0];
  if(IS_USED(used_mask, 1))
    sel1 = commands[1];
  if(IS_USED(used_mask, 2))
    sel2 = commands[2];

  /* select count mode */
  if(count_mode == PCL_MODE_USER)
    mode = 0x01;
  else if(count_mode == PCL_MODE_SYSTEM)
    mode = 0x02;
  else
    mode = 0x03;

  /* get actual processor cycle counter value (to measure elapsed cycles) */
  hw_descr->cycle_counter_value = ev5_read_cycle_counter();

  /* Clear and start the counters */
  return (ev5_perfctr_start(sel0, sel1, sel2, mode) == 0)
    ? PCL_SUCCESS
    : PCL_FAILURE;
}


/*                                                                            */
/*============================================================================*/
/* Read hardware counters                                                     */

int
PCL_driver_read(
		PCL_DESCR_TYPE d, /* I: descriptor */
		int max_counter,
		PCL_BIT_MASK_TYPE counter_used_mask,
		PCL_CNT_TYPE *counter_values
		)

{
  long read_cycles;
  PCL_BIT_MASK_TYPE used_mask;
  PCL_CNT_TYPE val0, val1, val2;
  PCL_state_descr *descr;
  HW_state_descr *hw_descr;


  /* get state descriptor */
  descr = d;
  hw_descr = &descr->hw_state_descr;

  /* set up used mask */
  used_mask = counter_used_mask & PCL_mask[max_counter-1];

  /* read out performance counters */
  if(ev5_perfctr_read(&val0, &val1, &val2) != 0)
    return PCL_FAILURE;

  if(IS_USED(used_mask, 0))
    counter_values[0] = val0;
  if(IS_USED(used_mask, 1))
    counter_values[1] = val1;
  if(IS_USED(used_mask, 2))
    counter_values[2] = val2;

  /* special case: cycle count */
  if(IS_USED(used_mask, 3))
    {
      read_cycles = ev5_read_cycle_counter();
      if(read_cycles < hw_descr->cycle_counter_value)
	/* overflow condition */
	counter_values[3] = read_cycles + ((1l << 32)
					   - hw_descr->cycle_counter_value);
      else
	counter_values[3] = read_cycles - hw_descr->cycle_counter_value;
    }

  /* everything ok */
  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/*                                                                            */


int
PCL_driver_stop(
		PCL_DESCR_TYPE d /* I: descriptor */
		)

{
  long ctr0, ctr1, ctr2;
  PCL_state_descr *descr;
  HW_state_descr *hw_descr;


  /* get state descriptor */
  descr = d;
  hw_descr = &descr->hw_state_descr;

#if defined(CHECK_MIGRATION)
  /* Check that we are still on the same processor. Eventually, the process
     was migrated to another processor. In that case we have wrong numbers.
  */
  if(hw_descr->processornumber != PCL_processor_number())
    return PCL_FAILURE;
#endif

  /* read out and stop performance counters */
  if(ev5_perfctr_stopread(&ctr0, &ctr1, &ctr2) != 0)
    return PCL_FAILURE;
  
  /* everything ok */
  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/* close                                                                      */

int
PCL_driver_close(
		 PCL_DESCR_TYPE d /* I: descriptor */
		 )

{
  /* everything ok */
  return PCL_SUCCESS;
}


/*============================================================================*
 *                             that's all folks                               *
 *============================================================================*/
