/*==============================================================================


   Purpose:    low level routines for MIPS


   Author:     Rudolf Berrendorf
               University of Applied Sciences Bonn-Rhein-Sieg
               Computer Science Department
               53754 Sankt Augustin
               rudolf.berrendorf@fh-bonn-rhein-sieg.de


   last mod:   23.08.2000


   DISCLAIMER

                  PCL - The Performance Counter Library


   Copyright (C) 2000
   Forschungszentrum Juelich GmbH, Federal Republic of Germany.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:

   Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

     - Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.

     - Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.

     - Any publications that result from the use of this software shall
       reasonably refer to the Research Centre's development.

     - All advertising materials mentioning features or use of this software
       must display the following acknowledgement:

           This product includes software developed by Forschungszentrum
           Juelich GmbH, Federal Republic of Germany.

     - Forschungszentrum Juelich GmbH is not obligated to provide the user with
       any support, consulting, training or assistance of any kind with regard
       to the use, operation and performance of this software or to provide
       the user with any updates, revisions or new versions.


   THIS SOFTWARE IS PROVIDED BY FORSCHUNGSZENTRUM JUELICH GMBH "AS IS" AND ANY
   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
   DISCLAIMED. IN NO EVENT SHALL FORSCHUNGSZENTRUM JUELICH GMBH BE LIABLE FOR
   ANY SPECIAL, DIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
   RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
   CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE ACCESS, USE OR PERFORMANCE OF THIS SOFTWARE.
  
==============================================================================*/

#include "../../defines.h"
#include "../../pcl.h"
#include "hw_driver_defs.h"
#include "../../pcl_hw_driver.h"
#include "../../hw_driver_global_defs.h"
#include "../../state_descr.h"

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/hwperfmacros.h>
#include <sys/hwperftypes.h>
#include <sys/procfs.h>
#include  <invent.h>


/*                                                                            */
/*----------------------------------------------------------------------------*/
/* local variables                                                            */

/* invalid entry-data if event is not available */
#define PCL_INVALID    0, {{{-1,-1},{0,0}}}

/* Entries are control codes for hardware performance counter.
   Values for the second counter start with 16.
*/
static const PCL_event_descr_t event_table_r10k[PCL_MAX_EVENT] =
{
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L1CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L1CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L1CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L1CACHE_HIT */
  {PCL_AVAIL,COMPOSED,{{{9,16+9},{OP_ADD,OP_ADD}}}}, /* PCL_L1CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L1DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L1DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L1DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L1DCACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,16+9},{-1,OP_ADD}}}}, /* PCL_L1DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                 /* PCL_L1ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L1ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L1ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L1ICACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{9,-1},{OP_ADD,-1}}}},  /* PCL_L1ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L2CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L2CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L2CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L2CACHE_HIT */
  {PCL_AVAIL,COMPOSED,{{{10,16+10},{OP_ADD,OP_ADD}}}}, /* PCL_L2CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L2DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L2DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L2DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L2DCACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,16+10},{-1,OP_ADD}}}}, /* PCL_L2DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L2ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L2ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L2ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L2ICACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{10,-1},{OP_ADD,-1}}}}, /* PCL_L2ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_TLB_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,16+7},{-1,OP_ADD}}}}, /* PCL_TLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_ITLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_ITLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_DTLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_DTLB_MISS */
  {PCL_AVAIL,NOT_COMP,{{{0,16+0},{OP_ADD,OP_ADD}}}}, /* PCL_CYCLES */
  {PCL_UNAVAIL, PCL_INVALID},                 /* PCL_ELAPSED_CYCLES */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_INTEGER_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{-1,16+5},{-1,OP_ADD}}}}, /* PCL_FP_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{-1,16+2},{-1,OP_ADD}}}}, /* PCL_LOAD_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{-1,16+3},{-1,OP_ADD}}}}, /* PCL_STORE_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_LOADSTORE_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{15,16+1},{OP_ADD,OP_ADD}}}}, /* PCL_INSTR */
  {PCL_AVAIL,COMPOSED,{{{6,16+8},{OP_ADD,OP_SUB}}}}, /* PCL_JUMP_SUCCESS */
  {PCL_AVAIL,NOT_COMP,{{{-1,16+8},{-1,OP_ADD}}}}, /* PCL_JUMP_UNSUCCESS */
  {PCL_AVAIL,NOT_COMP,{{{6,-1},{OP_ADD,-1}}}},  /* PCL_JUMP */
  {PCL_AVAIL,COMPOSED,{{{5,16+4},{OP_SUB,OP_ADD}}}}, /* PCL_ATOMIC_SUCCESS */
  {PCL_AVAIL,NOT_COMP,{{{5,-1},{OP_ADD,-1}}}},  /* PCL_ATOMIC_UNSUCCESS */
  {PCL_AVAIL,NOT_COMP,{{{-1,16+4},{OP_ADD,-1}}}}, /* PCL_ATOMIC */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_STALL_INTEGER */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_STALL_FP */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_STALL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_STALL_LOAD */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_STALL_STORE */
  {PCL_UNAVAIL, PCL_INVALID},                 /* PCL_STALL */
  {PCL_AVAIL,COMPOSED,{{{0,16+5},{OP_FP_1|OP_FP_2|OP_FP_REZ_MHZ,OP_FP_2|OP_FP_MUL}}}}, /* PCL_MFLOPS */
  {PCL_AVAIL,COMPOSED,{{{15,16},{OP_FP_2|OP_FP_STORE,OP_FP_2|OP_FP_DIV}}}}, /* PCL_IPC */
  {PCL_UNAVAIL, PCL_INVALID},                 /* PCL_L1DCACHE_MISSRATE */
  {PCL_UNAVAIL, PCL_INVALID},                 /* PCL_L2DCACHE_MISSRATE */
  {PCL_UNAVAIL, PCL_INVALID}                  /* PCL_MEM_FP_RATIO */
};

/* entries are control codes for hardware performance counter */
static const PCL_event_descr_t event_table_r12k[PCL_MAX_EVENT] =
{
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L1CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L1CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L1CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L1CACHE_HIT */
  {PCL_AVAIL,COMPOSED,{{{9,25},{OP_ADD,OP_ADD}}}}, /* PCL_L1CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L1DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L1DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L1DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L1DCACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,25},{-1,OP_ADD}}}},  /* PCL_L1DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                 /* PCL_L1ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L1ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L1ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L1ICACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{9,-1},{OP_ADD,OP_ADD}}}}, /* PCL_L1ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L2CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L2CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L2CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L2CACHE_HIT */
  {PCL_AVAIL,COMPOSED,{{{10,26},{OP_ADD,OP_ADD}}}}, /* PCL_L2CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L2DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L2DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L2DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L2DCACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,26},{-1,OP_ADD}}}}, /* PCL_L2DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L2ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L2ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L2ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L2ICACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{10,-1},{OP_ADD,-1}}}}, /* PCL_L2ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_TLB_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,23},{-1,OP_ADD}}}}, /* PCL_TLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_ITLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_ITLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_DTLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_DTLB_MISS */
  {PCL_AVAIL,NOT_COMP,{{{0,-1},{OP_ADD,-1}}}}, /* PCL_CYCLES */
  {PCL_UNAVAIL, PCL_INVALID},                 /* PCL_ELAPSED_CYCLES */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_INTEGER_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{-1,21},{-1,OP_ADD}}}}, /* PCL_FP_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{-1,18},{-1,OP_ADD}}}}, /* PCL_LOAD_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{-1,19},{-1,OP_ADD}}}}, /* PCL_STORE_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},                 /* PCL_LOADSTORE_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{15,-1},{OP_ADD,-1}}}}, /* PCL_INSTR */
  {PCL_AVAIL,COMPOSED,{{{6,24},{OP_ADD,OP_SUB}}}}, /* PCL_JUMP_SUCCESS */
  {PCL_AVAIL,NOT_COMP,{{{-1,24},{-1,OP_ADD}}}}, /* PCL_JUMP_UNSUCCESS */
  {PCL_AVAIL,NOT_COMP,{{{6,-1},{OP_ADD,-1}}}}, /* PCL_JUMP */
  {PCL_AVAIL,COMPOSED,{{{5,20},{OP_SUB,OP_ADD}}}}, /* PCL_ATOMIC_SUCCESS */
  {PCL_AVAIL,NOT_COMP,{{{5,-1},{OP_ADD,-1}}}}, /* PCL_ATOMIC_UNSUCCESS */
  {PCL_AVAIL,NOT_COMP,{{{-1,20},{-1,OP_ADD}}}}, /* PCL_ATOMIC */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_STALL_INTEGER */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_STALL_FP */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_STALL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_STALL_LOAD */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_STALL_STORE */
  {PCL_UNAVAIL, PCL_INVALID},                 /* PCL_STALL */
  {PCL_AVAIL,COMPOSED,{{{0,21},{OP_FP_1|OP_FP_2|OP_FP_REZ_MHZ,OP_FP_2|OP_FP_MUL}}}}, /* PCL_MFLOPS */
  {PCL_UNAVAIL, PCL_INVALID},                 /* PCL_IPC */
  {PCL_UNAVAIL, PCL_INVALID},                 /* PCL_L1DCACHE_MISSRATE */
  {PCL_UNAVAIL, PCL_INVALID},                 /* PCL_L2DCACHE_MISSRATE */
  {PCL_UNAVAIL, PCL_INVALID}                  /* PCL_MEM_FP_RATIO */
};

/*                                                                            */
/*============================================================================*/
/* initialization                                                             */

int
PCL_driver_open(
		PCL_DESCR_TYPE d /* I: descriptor */
		)

{
  char pfile[12];
  pid_t pid;
  int cntr, found;
  inventory_t *iv;
  inv_state_t *st;
  PCL_state_descr *descr = d;
  HW_state_descr *hw_descr;


  /* get state descriptor */
  descr = d;
  hw_descr = &descr->hw_state_descr;

  /* allowed count modes */
  /* ??? still correct? */
  descr->allowed_count_mode = PCL_MODE_USER;


  /* open the /proc-interface for performance counter access */
  if((pid = getpid()) == -1)
    return PCL_FAILURE;
  sprintf(pfile, "/proc/%05d", (int)pid);
  if((hw_descr->fd = open(pfile, O_RDWR)) == -1)
    return PCL_FAILURE;

  /* determine which type of CPU we are running on: go through inventory  */
  st = NULL;
  if(setinvent_r(&st) == -1)
    return PCL_FAILURE;
  found = 0;
  for(iv = getinvent_r(st); !found && (iv != NULL); iv = getinvent_r(st))
    {
      if((iv->inv_class == INV_PROCESSOR)
	 && (iv->inv_type == INV_CPUCHIP))
	{
	  found = 1;
	  /* This is a wild guess! Anybody has detailed information? */
	  if(iv->inv_state < 2000)
	    /* no R1x000 processor */
	    return PCL_FAILURE;
	  else if(iv->inv_state < 3000)
	    descr->event_table = event_table_r10k;
	  else
	    descr->event_table = event_table_r12k;
#if 0
	  printf("class=%d  type=%d contr=%d unit=%c state=%ld\n",
		 iv->inv_class, iv->inv_type, iv->inv_controller, iv->inv_unit,
		 iv->inv_state);
#endif
	}
    }
  endinvent_r(st);
  if(!found)
    return PCL_FAILURE;
  
  /* Now initialize a data structure used in counter descriptions.
     Ever seen such a few things set up in such a complicated interface?
  */
  hw_descr->evctr_args_init.hwp_ovflw_sig = 0;
  for(cntr = 0; cntr < HWPERF_EVENTMAX; ++cntr)
    {
      /* overflow frequency */
      hw_descr->evctr_args_init.hwp_ovflw_freq[cntr] = 0;
      /* don't use counter */
      hw_descr->evctr_args_init.hwp_evctrargs.hwp_evctrl[cntr].hwperf_spec = 0;
    }

  /* everything ok */
  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/* Starts hardware counters.                                                  */


int
PCL_driver_start(
		 PCL_DESCR_TYPE d, /* I: descriptor */
		 int max_counter,
		 PCL_BIT_MASK_TYPE counter_used_mask,
		 PCL_DRIVER_COMMAND_TYPE *commands,
		 unsigned int count_mode
		 )

{
  int sel0, sel1;
  hwperf_profevctrarg_t evctr_args;
  PCL_BIT_MASK_TYPE used_mask;
  PCL_state_descr *descr;
  HW_state_descr *hw_descr;


  /* get state descriptor */
  descr = d;
  hw_descr = &descr->hw_state_descr;


  /* check correct counting mode */
  if(count_mode != PCL_MODE_USER)
    return PCL_MODE_NOT_SUPPORTED;

  /* initialize counter descriptors */
  memcpy(&evctr_args, &hw_descr->evctr_args_init, sizeof(evctr_args));
    
  /* set up used mask */
  used_mask = counter_used_mask & PCL_mask[max_counter-1];

  if(IS_USED(used_mask, 0))
    {
      /* counter 0 used */
      sel0 = hw_descr->sel0 = commands[0];
      /* event to be counted */
      evctr_args.hwp_evctrargs.hwp_evctrl[sel0].hwperf_creg.hwp_ev = sel0;
      /* overflow interrupt enable */
      evctr_args.hwp_evctrargs.hwp_evctrl[sel0].hwperf_creg.hwp_ie = 1;
      /* user/kernel mode */
      evctr_args.hwp_evctrargs.hwp_evctrl[sel0].hwperf_creg.hwp_mode = HWPERF_CNTEN_U;
    }

  if(IS_USED(used_mask, 1))
    {
      /* counter 1 */
      sel1 = hw_descr->sel1 = commands[1];
      /* event to be counted */
      evctr_args.hwp_evctrargs.hwp_evctrl[sel1].hwperf_creg.hwp_ev = sel1 - 16;
      /* overflow interrupt enable */
      evctr_args.hwp_evctrargs.hwp_evctrl[sel1].hwperf_creg.hwp_ie = 1;
      /* user/kernel mode */
      evctr_args.hwp_evctrargs.hwp_evctrl[sel1].hwperf_creg.hwp_mode = HWPERF_CNTEN_U;
    }

  /* acquire counters */
  hw_descr->generation_number = ioctl(hw_descr->fd, PIOCENEVCTRS, (void *)&evctr_args);
  return (hw_descr->generation_number < 0) ? PCL_FAILURE : PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/* Read hardware counters                                                     */

int
PCL_driver_read(
		PCL_DESCR_TYPE d, /* I: descriptor */
		int max_counter,
		PCL_BIT_MASK_TYPE counter_used_mask,
		PCL_CNT_TYPE *counter_values
		)

{
  hwperf_cntr_t cnts;
  PCL_BIT_MASK_TYPE used_mask;
  PCL_state_descr *descr;
  HW_state_descr *hw_descr;


  /* get state descriptor */
  descr = d;
  hw_descr = &descr->hw_state_descr;

  /* set up used mask */
  used_mask = counter_used_mask & PCL_mask[max_counter-1];

  /* read hardware counters */
  if(ioctl(hw_descr->fd, PIOCGETEVCTRS, (void *)&cnts) != hw_descr->generation_number)
    return PCL_FAILURE;

  /* Extract selected events and store result values. */
  if(IS_USED(used_mask, 0))
    {
      counter_values[0] = cnts.hwp_evctr[hw_descr->sel0];
    }
  if(IS_USED(used_mask, 1))
    {
      counter_values[1] = cnts.hwp_evctr[hw_descr->sel1];
    }

  /* everything ok */
  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/*                                                                            */


int
PCL_driver_stop(
		PCL_DESCR_TYPE d /* I: descriptor */
		)

{
  PCL_state_descr *descr;
  HW_state_descr *hw_descr;


  /* get state descriptor */
  descr = d;
  hw_descr = &descr->hw_state_descr;

  if(ioctl(hw_descr->fd, PIOCRELEVCTRS) < 0)
    return PCL_FAILURE;

  /* everything ok */
  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/* close                                                                      */

int
PCL_driver_close(
		 PCL_DESCR_TYPE d /* I: descriptor */
		 )

{
  /* everything ok */
  return PCL_SUCCESS;
}


/*============================================================================*
 *                             that's all folks                               *
 *============================================================================*/
